/* ##########################################################################

    Copyright 2004-2009:
    ANTONIO JAVIER BARRAGAN, antonio.barragan@diesia.uhu.es
    http://www.uhu.es/antonio.barragan
    JOSE MANUEL ANDUJAR, andujar@diesia.uhu.es

    DPTO. DE ING. ELECTRONICA, DE SISTEMAS INFORMATICOS Y AUTOMATICA
    E.P.S. LA RABIDA, UNIVERSITY OF HUELVA (SPAIN)

    For more information, please contact with authors.

########################################################################## */

/*
    This software is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "aux_matlab.h"

#ifndef strcmpi
    #define strcmpi strcasecmp // Make the change to compile in GNU/Linux
#endif


#if defined(MATLAB_MEX_FILE) // These functions only are included if they are compiled in MATLAB

// FIS2System #############################################################

int FIS2System(const mxArray *FIS, System &S)
{
	mxArray *temp, *inputs, *outputs, *rules;
	char text[MAX_CHAR_LONG];
	int n,m,M_in,M_out,M_R,i,j;

	// Initial checks
	i=mxGetNumberOfFields(FIS);
	if (i<10 || i>13)
	{
		mexPrintf("%s\n",E_NoFIS);
		mexPrintf("%s\n\n",U_SeeManual);
		return 0;
	}
	temp=mxGetFieldByNumber(FIS,0,mxGetFieldNumber(FIS,"type")); // Type must be SUGENO
	if (mxGetString(temp,text,MAX_CHAR_LONG))
	{
		mexPrintf("%s\n",E_Model);
		mexPrintf("%s\n",E_NoFIS);
		mexPrintf("%s\n\n",U_SeeManual);
		return 0;
	}
	if (strcmpi(text,"SUGENO"))
	{
		mexPrintf("%s\n\n",E_NoSugeno);
		return 0;
	}
	temp=mxGetFieldByNumber(FIS,0,mxGetFieldNumber(FIS,"andMethod")); // Inference must be PROD
	if (mxGetString(temp,text,MAX_CHAR_LONG))
	{
		mexPrintf("%s.\n",E_Model);
		mexPrintf("%s\n",E_NoFIS);
		mexPrintf("%s\n\n",U_SeeManual);
		return 0;
	}
	if (strcmpi(text,"PROD"))
	{
		mexPrintf("%s\n\n",E_NoProd);
		return 0;
	}
	temp=mxGetFieldByNumber(FIS,0,mxGetFieldNumber(FIS,"defuzzMethod")); // Defuzzification method must be WTAVER
	if (mxGetString(temp,text,MAX_CHAR_LONG))
	{
		mexPrintf("%s\n",E_Model);
		mexPrintf("%s\n",E_NoFIS);
		mexPrintf("%s\n\n",U_SeeManual);
		return 0;
	}
	if (strcmpi(text,"WTAVER"))
	{
		mexPrintf("%s\n\n",E_NoSumProm);
		return 0;
	}

	// Read model parameters

	inputs=mxGetFieldByNumber(FIS,0,mxGetFieldNumber(FIS,"input")); // Inputs struct
	outputs=mxGetFieldByNumber(FIS,0,mxGetFieldNumber(FIS,"output")); // Outputs struct
	rules=mxGetFieldByNumber(FIS,0,mxGetFieldNumber(FIS,"rule")); // Rules struct

	M_in=mxGetN(inputs); // Number of inputs
	M_out=mxGetN(outputs); // Number of outputs
	M_R=mxGetN(rules); // Number of Rules (the sum of the rules of all outputs)

	if (M_out==0)
	{
		mexPrintf("%s\n\n",E_NoOutputs);
		mxDestroyArray(inputs);mxDestroyArray(outputs);mxDestroyArray(rules);
		return 0;
	}
	if (M_in==0)
	{
		mexPrintf("%s\n\n",E_NoInputs);
		mxDestroyArray(inputs);mxDestroyArray(outputs);mxDestroyArray(rules);
		return 0;
	}
	if (M_R<M_out)
	{
		mexPrintf("%s\n\n",E_NoRules);
		mxDestroyArray(inputs);mxDestroyArray(outputs);mxDestroyArray(rules);
		return 0;
	}

	// Create the model and read data
	int l,p,v,r;
	int v_antec, v_conseq, comprobar_conseq, error;
	int *num_rules=new int[M_out];
	for (i=0;i<M_out;i++)
		*(num_rules+i)=0;
	double *min_in=new double[M_in];
	double *max_in=new double[M_in];
	double *min_out=new double[M_out];
	double *max_out=new double[M_out];
	double d;
	bool constant, linear;
	char type[MAX_SIZE_TYPE_NAME];
	char TSKtype[MAX_SIZE_TYPE_NAME];
	mxArray *antecedent, *consequent, *mf_antec, *TSK;
	TYPE_MF T=NOMF;
	Membership *P;
	Rule *R;
	System Sist;

	// Count and check rules
	for (l=0;l<M_R;l++)
	{
		temp=mxGetFieldByNumber(rules,l,mxGetFieldNumber(rules,"weight")); // Weight must be 1
		if (mxGetScalar(temp)!=1)
		{
			mexPrintf("%s %d %s 1.\n\n",O_RuleWeigh,l+1,O_DifferentOf);
			mxDestroyArray(inputs);mxDestroyArray(outputs);mxDestroyArray(rules);
			delete [] num_rules; delete [] min_in; delete [] max_in; delete []min_out; delete [] max_out;
			return 0;
		}
		temp=mxGetFieldByNumber(rules,l,mxGetFieldNumber(rules,"connection")); // Connection must be 1
		if (mxGetScalar(temp)!=1)
		{
			mexPrintf("%s %d %s 1, AND.\n\n",O_RuleConnection,l+1,O_DifferentOf);
			mxDestroyArray(inputs);mxDestroyArray(outputs);mxDestroyArray(rules);
			delete [] num_rules; delete [] min_in; delete [] max_in; delete []min_out; delete [] max_out;
			return 0;
		}
		consequent=mxGetFieldByNumber(rules,l,mxGetFieldNumber(rules,"consequent")); // Get the consequent (struct)
		if (mxGetN(consequent)!=M_out)
		{
			mexPrintf("%s %d %s",O_ConsequentSize,l+1,O_IsNotCorrect);
			mxDestroyArray(inputs);mxDestroyArray(outputs);mxDestroyArray(rules);
			delete [] num_rules; delete [] min_in; delete [] max_in; delete []min_out; delete [] max_out;
			return 0;
		}
		comprobar_conseq=0; // Check that only 1 consequent for rule are defined
		for (i=0;i<M_out;i++)
		{
			v_conseq=(int)*(mxGetPr(consequent)+i);
			v_conseq--; // Number of consequent for the rule
			if (v_conseq>=0)
			{
				num_rules[i]++;
				if (comprobar_conseq==1)
				{
					mexPrintf("%s\n",E_No1Conseq);
					mxDestroyArray(inputs);mxDestroyArray(outputs);mxDestroyArray(rules);
					delete [] num_rules; delete [] min_in; delete [] max_in; delete []min_out; delete [] max_out;
					return 0;
				}
				comprobar_conseq=1;
			}
		}
	}
	// Create the System
	Sist.initialize(M_in,M_out,num_rules);

	// Use 'num_rules' to index the rules of each output
	mxArray * limites;
	for (j=0;j<M_out;j++)
	{
		*(num_rules+j)=0;
		limites=mxGetFieldByNumber(outputs,j,mxGetFieldNumber(outputs,"range")); // Get j-output range
		min_out[j]=*mxGetPr(limites);
		max_out[j]=*(mxGetPr(limites)+1);
}

	for (i=0;i<M_in;i++)
	{
		limites=mxGetFieldByNumber(inputs,i,mxGetFieldNumber(inputs,"range"));// Get i-input range
		min_in[i]=*mxGetPr(limites);
		max_in[i]=*(mxGetPr(limites)+1);
	}

	Sist.in_min(min_in);
	Sist.in_max(max_in);
	Sist.out_min(min_out);
	Sist.out_max(max_out);

	delete [] min_in; delete [] max_in; delete []min_out; delete [] max_out;

	for (l=0;l<M_R;l++)
	{
		antecedent=mxGetFieldByNumber(rules,l,mxGetFieldNumber(rules,"antecedent")); // Get the antecedent (struct)
		if (mxGetN(antecedent)!=M_in)
		{
			mexPrintf("%s %d %s",O_AntecedentSize,l+1,O_IsNotCorrect);
			mxDestroyArray(inputs);mxDestroyArray(outputs);mxDestroyArray(rules);
			mxDestroyArray(antecedent);
			delete [] num_rules;
			return 0;
		}
		consequent=mxGetFieldByNumber(rules,l,mxGetFieldNumber(rules,"consequent")); // Get the consequent (struct)
		for (i=0;i<M_out;i++)
		{
			v_conseq=(int)*(mxGetPr(consequent)+i);
			v_conseq--; // Number of consequent for the rule
			if (v_conseq>=0)
			{
				TSK=mxGetFieldByNumber(outputs,i,mxGetFieldNumber(outputs,"mf")); // Get the consequent of the output 'i' (struct) // Mf
				if (mxGetString(mxGetFieldByNumber(TSK,v_conseq,mxGetFieldNumber(TSK,"type")),TSKtype,MAX_SIZE_TYPE_NAME)) // Consecuent type (TSK or Singleton)
				{
					mexPrintf("%s\n",E_Model);
					mexPrintf("%s %d.\n\n",E_Conseq,l+1);
					mxDestroyArray(inputs);mxDestroyArray(outputs);mxDestroyArray(rules);
					mxDestroyArray(antecedent);mxDestroyArray(consequent);
					delete [] num_rules;
					return 0;
				}
				constant=!strcmpi(TSKtype,"CONSTANT");
				linear=!strcmpi(TSKtype,"LINEAR");
				for (p=0;p<M_in;p++)
				{
					R=Sist.read_Rule(num_rules[i],i);
					if (R==NULL)
					{
						mexPrintf("%s\n",E_BadModel);
						mxDestroyArray(inputs);mxDestroyArray(outputs);mxDestroyArray(rules);
						mxDestroyArray(antecedent);mxDestroyArray(consequent);
						delete [] num_rules;
						return 0;
					}
					v_antec=(int)*(mxGetPr(antecedent)+p);
					v_antec--; // Number of antecedent for the rule
					mf_antec=mxGetFieldByNumber(inputs,p,mxGetFieldNumber(inputs,"mf")); // Get the antecedent of the input 'p' (struct) // Mf
					if (mxGetString(mxGetFieldByNumber(mf_antec,v_antec,mxGetFieldNumber(mf_antec,"type")),type,MAX_SIZE_TYPE_NAME)) // Read the membership function type
					{
						mexPrintf("%s\n",E_Model);
						mexPrintf("%s %d, %s %d.\n\n",E_MFType,l+1,O_Input,p+1);
						mxDestroyArray(inputs);mxDestroyArray(outputs);mxDestroyArray(rules);
						mxDestroyArray(antecedent);mxDestroyArray(consequent);
						delete [] num_rules;
						return 0;
					}
					for (int iT=0;iT<=(int)ZMF;iT++) // Check if the MF type
					{
						T=(TYPE_MF)iT;
						if (!strcmpi(MF_NAMES[T],type))
							break;
					}
					if (T>ZMF)
					{
						mexPrintf("%s %s %d, %s %d.\n\n",E_BadMF,O_Rule,l+1,O_Input,p+1);
						mxDestroyArray(inputs);mxDestroyArray(outputs);mxDestroyArray(rules);
						mxDestroyArray(antecedent);mxDestroyArray(consequent);
						delete [] num_rules;
						return 0;
					}
					if (MF_PARAM_NUMBER[(int)T]!=mxGetN(mxGetFieldByNumber(mf_antec,v_antec,mxGetFieldNumber(mf_antec,"params")))) 
					{
						mexPrintf("%s %s %d. %s %d.\n\n",E_NumberParam,O_MF,p+1,O_Rule,l+1);
						mxDestroyArray(inputs);mxDestroyArray(outputs);mxDestroyArray(rules);
						mxDestroyArray(antecedent);mxDestroyArray(consequent);
						delete [] num_rules;
						return 0;
					}
					if (!R->change_function(T,p))
					{
						mexPrintf("%s %s %d, %s %d.\n\n",E_MFTypeDef,l+1,p+1,O_Rule,O_Input);
						mxDestroyArray(inputs);mxDestroyArray(outputs);mxDestroyArray(rules);
						mxDestroyArray(antecedent);mxDestroyArray(consequent);
						delete [] num_rules;
						return 0;
					}
					P=createMF(T);
					for (v=0;v<P->n_param();v++)
					{
						d=*(mxGetPr(mxGetFieldByNumber(mf_antec,v_antec,mxGetFieldNumber(mf_antec,"params")))+v);
						// Change the parameters for use with MATLAB (change the definition of Gauss functions)
						if (T==GAUSSMF || T==GAUSS2MF)
						{
							if (v==0 || v==2)
							{
								d=M_SQRT2*d;
								error=P->edit(v+1,d);
							}
							else if (v==1 || v==3)
								error=P->edit(v-1,d);
						}
						else
							error=P->edit(v,d);
						if (!error || P->test()==0)
						{
							mexPrintf("%s\n\n",E_ParamMF);
							mxDestroyArray(inputs);mxDestroyArray(outputs);mxDestroyArray(rules);
							mxDestroyArray(antecedent);mxDestroyArray(consequent);
							delete [] num_rules;
							delete P;
							P=NULL;
							return 0;
						}
					}
					if(!R->write_function(*P,p))
					{
						mexPrintf("%s\n\n",E_MFDef);
						mxDestroyArray(inputs);mxDestroyArray(outputs);mxDestroyArray(rules);
						mxDestroyArray(antecedent);mxDestroyArray(consequent);
						delete [] num_rules;
						delete P;
						P=NULL;
						return 0;
					}
					delete P;
					P=NULL;
					for (v=0;v<=M_in;v++)
					{
						if(linear)
						{
							d=*(mxGetPr(mxGetFieldByNumber(TSK,v_conseq,mxGetFieldNumber(TSK,"params")))+v);
							if (v!=M_in)
							{
								if(!R->write_TSK(d,v+1))
								{
									mexPrintf("%s\n\n",E_ParamConseq);
									mxDestroyArray(inputs);mxDestroyArray(outputs);mxDestroyArray(rules);
									mxDestroyArray(antecedent);mxDestroyArray(consequent);
									delete [] num_rules;
									return 0;
								}
							}
							else
							{
								if(!R->write_TSK(d,0))
								{
									mexPrintf("%s\n\n",E_ParamConseq);
									mxDestroyArray(inputs);mxDestroyArray(outputs);mxDestroyArray(rules);
									mxDestroyArray(antecedent);mxDestroyArray(consequent);
									delete [] num_rules;
									return 0;
								}
							}
						}
						else // constant
						{
							if (v!=0)
							{
								if(!R->write_TSK(0.0,v))
								{
									mexPrintf("%s\n\n",E_ParamConseq);
									mxDestroyArray(inputs);mxDestroyArray(outputs);mxDestroyArray(rules);
									mxDestroyArray(antecedent);mxDestroyArray(consequent);
									delete [] num_rules;
									return 0;
								}
							}
							else
							{
								d=*(mxGetPr(mxGetFieldByNumber(TSK,v_conseq,mxGetFieldNumber(TSK,"params")))+v);
								if(!R->write_TSK(d,0))
								{
									mexPrintf("%s\n\n",E_ParamConseq);
									mxDestroyArray(inputs);mxDestroyArray(outputs);mxDestroyArray(rules);
									mxDestroyArray(antecedent);mxDestroyArray(consequent);
									delete [] num_rules;
									return 0;
								}
							}
						}
					}
				}
				num_rules[i]++;
			}
		}
	}
	delete [] num_rules;
	S=Sist;
	return 1;
}

// System2FIS #############################################################
int System2FIS(System &S, mxArray *FIS)
{

	int i,j,r,p,index,L;
	const int in=S.inputs(),out=S.outputs();
	int *M=new int[out];
	Membership *P;
	Rule *R;
	L=0;
	for (i=0;i<out;i++)
	{
		M[i]=S.rules(i);
		L+=M[i];
		if (M[i]<1)
		{
			mexPrintf("%s\n",E_NoRules);
			delete [] M;
			return 0;
		}
	}
	char type[MAX_CHAR_LONG];

	if (out<1)
	{
		mexPrintf("%s\n",E_NoOutputs);
		delete [] M;
		return 0;
	}
	if (in<1)
	{
		mexPrintf("%s\n",E_NoInputs);
		delete [] M;
		return 0;
	}

	// Create the FIS struct
	mxSetFieldByNumber(FIS,0,0,mxCreateString(U_FISName));
	mxSetFieldByNumber(FIS,0,1,mxCreateString("sugeno"));
	mxSetFieldByNumber(FIS,0,2,mxCreateString("prod"));
	mxSetFieldByNumber(FIS,0,3,mxCreateString("probor"));
	mxSetFieldByNumber(FIS,0,4,mxCreateString("wtaver"));
	mxSetFieldByNumber(FIS,0,5,mxCreateString("prod"));
	mxSetFieldByNumber(FIS,0,6,mxCreateString("max"));

	// Create the input struct
	int dim[]={1,in};
	const char* InOutNames[]={"name","range","mf"};
	const char* mfNames[]={"name","type","params"};
	mxArray *inputs=mxCreateStructArray(2,dim,3,InOutNames);
	char char_num[MAX_CHAR_LONG], inputName[MAX_CHAR_LONG], nom_mf[MAX_CHAR_LONG];
	mxArray *range, *mf, *param;
	mxArray *outputs;
	double d;
	TYPE_MF temp;

	for (i=0;i<in;i++)
	{
		sprintf(&inputName[0],"%s%d",U_FISInput,i+1);
		mxSetFieldByNumber(inputs,i,0,mxCreateString(inputName));
		range=mxCreateDoubleMatrix(1,2,mxREAL);
		*(mxGetPr(range))=S.in_min(i);
		*(mxGetPr(range)+1)=S.in_max(i);
		mxSetFieldByNumber(inputs,i,1,range);
		dim[1]=L;
		mf=mxCreateStructArray(2,dim,3,mfNames);
		index=0;
		for (j=0;j<out;j++)
		{
			for (r=0;r<M[j];r++,index++)
			{
				R=S.read_Rule(r,j);
				if (R==NULL)
				{
					delete [] M;
					return 0;
				}
				P=R->read_function(i);
				if (P==NULL || P->test()==0)
				{
					delete [] M;
					return 0;
				}
				sprintf(nom_mf,"%s%d%s%d%s%d",U_RuleAbbreviation,r+1,U_InputAbbreviation,i+1,U_OutputAbreviation,j+1);
				mxSetFieldByNumber(mf,index,0,mxCreateString(nom_mf));
				temp=P->type();
				for (p=0;p<=strlen(MF_NAMES[(int)temp]);p++)
					type[p]=tolower(MF_NAMES[(int)temp][p]);
				mxSetFieldByNumber(mf,index,1,mxCreateString(type));
				param=mxCreateDoubleMatrix(1,P->n_param(),mxREAL); // Do not use MF_PARAM_NUMBER[(int)temp], because P->n_param() are more general
				for (p=0;p<P->n_param();p++)
					*(mxGetPr(param)+p)=P->read(p);

				// Change the parameters for use with MATLAB (change the definition of Gauss functions)
				if (temp==GAUSSMF || temp==GAUSS2MF)
				{
					d=mxGetScalar(param);
					(*(mxGetPr(param)))=(*(mxGetPr(param)+1))/M_SQRT2;
					*(mxGetPr(param)+1)=d;
				}
				if (temp==GAUSS2MF)
				{
					d=*(mxGetPr(param)+2);
					(*(mxGetPr(param)+2))=(*(mxGetPr(param)+3))/M_SQRT2;
					*(mxGetPr(param)+3)=d;
				}
				mxSetFieldByNumber(mf,index,2,param);
			}
		}
		mxSetFieldByNumber(inputs,i,2,mf);
	}
	mxSetFieldByNumber(FIS,0,7,inputs);

	// Create the output struct
	dim[1]=out;
	outputs=mxCreateStructArray(2,dim,3,InOutNames);
	char outputName[MAX_CHAR_LONG];
	for (j=0;j<out;j++)
	{
		sprintf(&outputName[0],"%s%d",U_FISOutput,j+1);
		mxSetFieldByNumber(outputs,j,0,mxCreateString(outputName));
		range=mxCreateDoubleMatrix(1,2,mxREAL);
		*(mxGetPr(range))=S.out_min(j);
		*(mxGetPr(range)+1)=S.out_max(j);
		mxSetFieldByNumber(outputs,j,1,range);
		dim[1]=M[j];
		mf=mxCreateStructArray(2,dim,3,mfNames);
		for (r=0;r<M[j];r++)
		{
			R=S.read_Rule(r,j);
			sprintf(&nom_mf[0],"%s%d%s%d",U_RuleAbbreviation,r+1,U_OutputAbreviation,j+1);
			mxSetFieldByNumber(mf,r,0,mxCreateString(nom_mf));
			mxSetFieldByNumber(mf,r,1,mxCreateString("linear"));
			param=mxCreateDoubleMatrix(1,in+1,mxREAL);
			for (i=0;i<in;i++)
				*(mxGetPr(param)+i)=R->read_TSK(i+1);
			*(mxGetPr(param)+in)=R->read_TSK(0);
			mxSetFieldByNumber(mf,r,2,param);
		}
		mxSetFieldByNumber(outputs,j,2,mf);
	}
	mxSetFieldByNumber(FIS,0,8,outputs);

	// Create the rule struct
	dim[1]=L;
	const char* ruleNames[]={"antecedent","consequent","weight","connection"};
	mxArray *rules=mxCreateStructArray(2,dim,4,ruleNames);
	mxArray *antecedent, *consequent;
	int input=0,output=0,value;
	index=0;
	for (j=0;j<out;j++)
	{
		for (r=0;r<M[j];r++,index++)
		{
			antecedent=mxCreateDoubleMatrix(1,in,mxREAL);
			mxSetFieldByNumber(rules,index,0,antecedent);
			for (i=0;i<in;i++)
				*(mxGetPr(antecedent)+i)=index+1;
			consequent=mxCreateDoubleMatrix(1,out,mxREAL);
			*(mxGetPr(consequent)+j)=(r+1);
			mxSetFieldByNumber(rules,index,1,consequent);
			mxSetFieldByNumber(rules,index,2,mxCreateDoubleScalar(1)); // Weight=1
			mxSetFieldByNumber(rules,index,3,mxCreateDoubleScalar(1)); // Connection=1
		}
	}
	mxSetFieldByNumber(FIS,0,9,rules);
	delete [] M;
	return 1;
}

// readModel ##############################################################

int readModel(const mxArray *model, System &S)
{
	if (mxIsChar(model))
	{
		char nombre[MAX_CHAR_LONG], extension[4];
		bool fin=0;
		int i,j;
		// The model is TXT or FIS file
		if (mxGetString(model,nombre,MAX_CHAR_LONG))
			return 0;
		for (i=0;(i<MAX_CHAR_LONG && fin==0);i++)
		{
			if(nombre[i]=='\0')
			{
				i=i-3;
				for (j=0;j<4;j++)
					extension[j]=nombre[i+j];
				fin=1;
			}
		}
		if (fin==0)
			return 0;
		if(!strcmpi(extension,"FIS"))
		{// Is FIS file
			mxArray *arg_der[1], *arg_izq[1];
			arg_der[0]=mxCreateString(nombre);
			if(mexCallMATLAB(1,arg_izq,1,arg_der,"readfis"))
			{
				mxDestroyArray(arg_der[0]);
				return 0;
			}
			if(!FIS2System(arg_izq[0],S))
			{
				mxDestroyArray(arg_der[0]);
				return 0;
			}
			mxDestroyArray(arg_izq[0]);
			mxDestroyArray(arg_der[0]);
		}
		else if(!strcmpi(extension,"TXT"))
		{// Is TXT file
			if (!TXT2System(&nombre[0],S))
				return 0;
		}
		else
			return 0;
	}
	else
	{
		// Is a FIS variable
		if (!mxIsStruct(model) || (mxGetNumberOfFields(model)<10 && mxGetNumberOfFields(model)>13))
			return 0;
		if(!FIS2System(model,S))
			return 0;
	}
	return 1;
}
#endif
